"use strict";

var Promise = require("bluebird");

var logger = require('../../logger');

var Alias = require('./Alias');
var IdentifiedElement = require('./IdentifiedElement');
var LayoutData = require('./LayoutData');
var PointData = require('./PointData');
var Reaction = require('./Reaction');

// This file describes javascript representation of Java Model class that
// represents content of the map.

/**
 * Default constructor.
 * 
 */

function MapModel(configuration) {

  // list of aliases is empty (it will be filled dynamically - when necessary)
  this._aliases = [];

  // list of reactions is empty (it will be filled dynamically - when necessary)
  this._reactions = [];
  this._reactionsByParticipantElementId = [];

  // list of aliases that should be updated from server side during the next
  // connection
  this._missingAliases = [];

  // list of reactions that should be updated from server side during the next
  // connection
  this._missingReactions = [];

  // list of layouts is empty (it will be filled when necessary)
  this._layoutsData = [];

  // information about points and associated data (for now we have only comments
  // associated to the point,
  // but it can be extended)
  this._pointsData = [];

  this._submodels = [];

  if (configuration !== undefined) {
    if (configuration instanceof MapModel) {
      this.setId(configuration.getId());
      this.setName(configuration.getName());
      this.setTileSize(configuration.getTileSize());
      this.setWidth(configuration.getWidth());
      this.setHeight(configuration.getHeight());
      this.setMinZoom(configuration.getMinZoom());
      this.setMaxZoom(configuration.getMaxZoom());
      this.addLayouts(configuration.getLayouts());
      this.addSubmodels(configuration.getSubmodels());
      this.setCenterLatLng(configuration.getCenterLatLng());
      this.setTopLeftLatLng(configuration.getTopLeftLatLng());
      this.setBottomRightLatLng(configuration.getBottomRightLatLng());
      this.setSubmodelType(configuration.getSubmodelType());
    } else {
      this.setId(configuration.idObject);
      this.setName(configuration.name);
      this.setTileSize(configuration.tileSize);
      this.setWidth(configuration.width);
      this.setHeight(configuration.height);
      this.setMinZoom(configuration.minZoom);
      this.setMaxZoom(configuration.maxZoom);
      this.addLayouts(configuration.layouts);
      this.addSubmodels(configuration.submodels);
      this.setCenterLatLng(configuration.centerLatLng);
      this.setTopLeftLatLng(configuration.topLeftLatLng);
      this.setBottomRightLatLng(configuration.bottomRightLatLng);
      this.setSubmodelType(configuration.submodelType);
    }
  }
}

/**
 * Returns list of {@link LayoutData} on this model.
 * 
 * @returns {Array} with list of {@link LayoutData} on this model
 */
MapModel.prototype.getLayoutsData = function() {
  var result = [];
  for ( var id in this._layoutsData) {
    if (this._layoutsData.hasOwnProperty(id)) {
      result.push(this._layoutsData[id]);
    }
  }
  return result;
};

MapModel.prototype.getLayouts = function() {
  return this.getLayoutsData();
};

/**
 * Return list of all aliases that were added to the model.
 */
MapModel.prototype.getAliases = function(params) {
  var self = this;
  return ServerConnector.getAliases({
    columns : "id,modelId",
    type : params.type,
    modelId : self.getId(),
    includedCompartmentIds : params.includedCompartmentIds,
    excludedCompartmentIds : params.excludedCompartmentIds,

  }).then(function(lightAliases) {
    var identifiedElements = [];
    for (var i = 0; i < lightAliases.length; i++) {
      self.addAlias(lightAliases[i]);
      identifiedElements.push(new IdentifiedElement(lightAliases[i]));
    }
    return self.getByIdentifiedElements(identifiedElements, params.complete);
  });
};

/**
 * Returns {@link Alias} by identifier.
 * 
 * @param id
 *          identifier of the {@link Alias}
 * @returns {@link Alias} by identifier
 */
MapModel.prototype.getAliasById = function(id, complete) {
  var self = this;
  if (complete) {
    return this.getCompleteAliasById(id);
  }
  if (self._aliases[id] !== undefined) {
    return Promise.resolve(self._aliases[id]);
  } else {
    return self.getMissingElements({
      aliasIds : [ id ]
    }).then(function() {
      return self._aliases[id];
    });
  }
};

MapModel.prototype.getCompleteAliasById = function(id) {
  var self = this;
  if (self._aliases[id] !== undefined && self._aliases[id].isComplete()) {
    return Promise.resolve(self._aliases[id]);
  } else {
    return ServerConnector.getAliases({
      ids : id
    }).then(function(aliases) {
      if (self._aliases[id] === undefined) {
        self._aliases[id] = aliases[0];
      } else {
        self._aliases[id].update(aliases[0]);
      }
      return self._aliases[id];
    });
  }
};

/**
 * Returns {@link Reaction} by identifier.
 * 
 * @param id
 *          identifier of the {@link Reaction}
 * @returns {@link Reaction} by identifier
 */
MapModel.prototype.getReactionById = function(id, complete) {
  var self = this;
  if (complete) {
    return this.getCompleteReactionById(id);
  }
  if (self._reactions[id] !== undefined) {
    return Promise.resolve(self._reactions[id]);
  } else {
    return self.getMissingElements({
      reactionIds : [ id ]
    }).then(function() {
      return self._reactions[id];
    });
  }
};

MapModel.prototype._getMissingReactionsElementIds = function(reactions) {
  var self = this;
  var result = [];
  var ids = [];
  for (var i = 0; i < reactions.length; i++) {
    var reaction = reactions[i];
    var elements = reaction.getElements();
    for (var j = 0; j < elements.length; j++) {
      var element = elements[j];
      if (!(element instanceof Alias)) {
        if (self._aliases[element] === undefined || !self._aliases[element].isComplete()) {
          if (!ids[element]) {
            ids[element] = true;
            result.push(element);
          }
        }
      }
    }
  }
  return result;
};

MapModel.prototype.getCompleteReactionById = function(id) {
  var self = this;
  if (self._reactions[id] instanceof Reaction && self._reactions[id].isComplete()) {
    return Promise.resolve(self._reactions[id]);
  } else {
    var result;
    return self.getReactionById(id).then(function(result) {
      var ids = self._getMissingReactionsElementIds([ result ]);
      return self.getMissingElements({
        aliasIds : ids,
        complete : true
      });
    }).then(function() {
      var i;
      result = self._reactions[id];
      for (i = 0; i < result.getReactants().length; i++) {
        if (!(result.getReactants()[i] instanceof Alias)) {
          result.getReactants()[i] = self._aliases[result.getReactants()[i]];
        }
      }
      for (i = 0; i < result.getProducts().length; i++) {
        if (!(result.getProducts()[i] instanceof Alias)) {
          result.getProducts()[i] = self._aliases[result.getProducts()[i]];
        }
      }
      for (i = 0; i < result.getModifiers().length; i++) {
        if (!(result.getModifiers()[i] instanceof Alias)) {
          result.getModifiers()[i] = self._aliases[result.getModifiers()[i]];
        }
      }
      return result;
    });
  }
};

MapModel.prototype.getMissingElements = function(elements) {
  var self = this;

  var layouts = this._getLayouts();
  var aliasIds = [];
  var reactionIds = [];

  var i = 0;
  if (elements.reactionIds !== undefined) {
    reactionIds.push.apply(reactionIds, elements.reactionIds);
    for (i = 0; i < reactionIds.length; i++) {
      this._missingReactions[reactionIds[i]] = reactionIds[i];
    }

  }
  if (elements.aliasIds !== undefined) {
    aliasIds.push.apply(aliasIds, elements.aliasIds);
    for (i = 0; i < aliasIds.length; i++) {
      this._missingAliases[aliasIds[i]] = aliasIds[i];
    }
  }

  for (i = 0; i < layouts.length; i++) {
    var layout = layouts[i];
    if (layout.isInitialized()) {
      var aliases = layout.getAliases();
      for (var j = 0; j < aliases.length; j++) {
        var alias = aliases[j];
        if (this._aliases[alias.getId()] === undefined && this._missingAliases[alias.getId()] === undefined) {
          this._missingAliases[alias.getId()] = alias.getId();
          aliasIds.push(alias.getId());
        }
      }

      var reactions = layout.getReactions();
      for (var k = 0; k < reactions.length; k++) {
        var reaction = reactions[k];
        if (this._reactions[reaction.getId()] === undefined && this._missingReactions[reaction.getId()] === undefined) {
          this._missingReactions[reaction.getId()] = reaction.getId();
          reactionIds.push(reaction.getId());
        }
      }
    }
  }
  var reactionPromise = null;
  if (reactionIds.length > 0) {
    reactionPromise = ServerConnector.getReactions({
      ids : reactionIds,
      complete : elements.complete
    });
  }

  var aliasPromise = null;
  if (aliasIds.length > 0) {
    if (elements.complete) {
      aliasPromise = ServerConnector.getAliases({
        ids : aliasIds
      });
    } else {
      aliasPromise = ServerConnector.getAliases({
        ids : aliasIds,
        columns : "id,bounds,modelId"
      });

    }
  }

  var result = [];
  return Promise.all([ reactionPromise, aliasPromise ]).then(function(values) {
    var i;
    var reactions = values[0];
    var aliases = values[1];
    var ids = [];

    if (reactions !== null) {
      for (i = 0; i < reactions.length; i++) {
        var reaction = reactions[i];
        self.addReaction(reaction);
        result.push(reaction);
      }
      ids = self._getMissingReactionsElementIds(reactions);
    }
    if (aliases !== null) {
      for (i = 0; i < aliases.length; i++) {
        var alias = aliases[i];
        self.addAlias(alias);
        result.push(alias);
      }
    }
    if (ids.length > 0) {
      return self.getMissingElements({
        aliasIds : ids,
        complete : true
      });
    } else {
      return Promise.resolve([]);
    }
  }).then(function() {
    return result;
  });
};

/**
 * Returns layout data for a given layout identifier.
 * 
 * @param layoutId
 *          layout identifier
 * @returns {LayoutData} for a given layout identifier
 */
MapModel.prototype.getLayoutDataById = function(layoutId) {
  var self = this;
  if (self._layoutsData[layoutId] !== undefined) {
    return Promise.resolve(self._layoutsData[layoutId]);
  } else {
    return ServerConnector.getOverlayById(layoutId).then(function(layout) {
      self.addLayout(layout);
      return self._layoutsData[layoutId];
    });
  }
};

/**
 * Adds information about alias.
 * 
 * @param aliasData
 *          raw data about alias
 */
MapModel.prototype.addAlias = function(aliasData) {
  var alias = aliasData;
  if (!(aliasData instanceof Alias)) {
    alias = new Alias(aliasData);
  }
  if (this._aliases[alias.getId()] !== undefined) {
    this._aliases[alias.getId()].update(alias);
  } else {
    this._aliases[alias.getId()] = alias;
    if (this._missingAliases[alias.getId()] !== undefined) {
      this._missingAliases[alias.getId()] = null;
      delete this._missingAliases[alias.getId()];
    }
  }
};

/**
 * Adds information about reaction.
 * 
 * @param reactionData
 *          raw data about reaction
 */
MapModel.prototype.addReaction = function(reactionData) {
  var reaction = null;
  if (reactionData instanceof Reaction) {
    reaction = reactionData;
  } else {
    reaction = new Reaction(reactionData);
  }
  if (this._reactions[reaction.getId()] !== undefined) {
    logger.warn("Reaction with id: " + reaction.getId() + " already exists");
  } else {
    this._reactions[reaction.getId()] = reaction;
    if (this._missingReactions[reaction.getId()] !== undefined) {
      this._missingReactions[reaction.getId()] = null;
      delete this._missingReactions[reaction.getId()];
    }
  }
};

/**
 * Returns {@link PointData} for a given point on the map.
 * 
 * @param point
 *          {@link google.maps.Point} where we are requesting data
 * @returns {@link PointData} for a given point on the map
 */
MapModel.prototype.getPointDataByPoint = function(inputPoint) {
  if (inputPoint instanceof google.maps.Point) {
    var point = this._roundPoint(inputPoint);
    var id = this._pointToId(point);
    var result = this._pointsData[id];
    if (result === undefined) {
      result = new PointData(point, this.getId());
      this._pointsData[id] = result;
    }
    return result;
  } else {
    logger.warn("point must be of class: google.maps.Point");
    return null;
  }
};

/**
 * Returns point where x and y coordinate are rounded to 2 decimal places.
 * 
 * @param point
 *          input point
 * @returns {google.maps.Point} point where x and y coordinate are rounded to 2
 *          decimal places
 */
MapModel.prototype._roundPoint = function(point) {
  var x = parseFloat(point.x).toFixed(2);
  var y = parseFloat(point.y).toFixed(2);
  return new google.maps.Point(x, y);
};

/**
 * Transform point into string identifier.
 * 
 * @param point
 *          {google.maps.Point} to transform
 * @returns {String} string identifier for a given point
 */
MapModel.prototype._pointToId = function(point) {
  if (point instanceof google.maps.Point) {
    return "(" + point.x + ", " + point.y + ")";
  } else {
    return point.replace(/ /g, '');
  }
};

MapModel.prototype.getId = function() {
  return this.id;
};

MapModel.prototype.setId = function(id) {
  this.id = parseInt(id);
};

MapModel.prototype.getWidth = function() {
  return this._width;
};

MapModel.prototype.setWidth = function(width) {
  this._width = width;
};

MapModel.prototype.getHeight = function() {
  return this._height;
};

MapModel.prototype.setHeight = function(height) {
  this._height = height;
};

MapModel.prototype.getName = function() {
  return this._name;
};

MapModel.prototype.setName = function(name) {
  this._name = name;
};

MapModel.prototype.getMinZoom = function() {
  return this._minZoom;
};

MapModel.prototype.setMinZoom = function(minZoom) {
  this._minZoom = minZoom;
};

MapModel.prototype.getSubmodelType = function() {
  return this._submodelType;
};

MapModel.prototype.setSubmodelType = function(submodelType) {
  this._submodelType = submodelType;
};

function createLatLng(param) {
  if (param === undefined) {
    return null;
  }
  if (param instanceof google.maps.LatLng) {
    return new google.maps.LatLng(param.lat(), param.lng());
  } else {
    return new google.maps.LatLng(param.lat, param.lng);
  }
}

MapModel.prototype.setCenterLatLng = function(centerLatLng) {
  var newVal = createLatLng(centerLatLng);
  if (newVal === null) {
    logger.warn("centerLatLng is invalid");
  } else {
    this._centerLatLng = newVal;
  }
};

MapModel.prototype.getCenterLatLng = function() {
  return this._centerLatLng;
};

MapModel.prototype.setTopLeftLatLng = function(topLeftLatLng) {
  var newVal = createLatLng(topLeftLatLng);
  if (newVal === null) {
    logger.warn("topLeftLatLng is invalid");
  } else {
    this._topLeftLatLng = newVal;
  }
};

MapModel.prototype.getTopLeftLatLng = function() {
  return this._topLeftLatLng;
};

MapModel.prototype.setBottomRightLatLng = function(bottomRightLatLng) {
  var newVal = createLatLng(bottomRightLatLng);
  if (newVal === null) {
    logger.warn("bottomRightLatLng is invalid");
  } else {
    this._bottomRightLatLng = newVal;
  }
};

MapModel.prototype.getBottomRightLatLng = function() {
  return this._bottomRightLatLng;
};

MapModel.prototype.getMaxZoom = function() {
  return this._maxZoom;
};

MapModel.prototype.setMaxZoom = function(maxZoom) {
  this._maxZoom = maxZoom;
};

MapModel.prototype.getTileSize = function() {
  return this._tileSize;
};

MapModel.prototype.getPictureSize = function() {
  return Math.max(this.getWidth(), this.getHeight());
};

MapModel.prototype.setTileSize = function(tileSize) {
  this._tileSize = tileSize;
};

MapModel.prototype.addLayouts = function(layouts) {
  if (layouts === undefined) {
    logger.warn("Layouts are undefined...");
  } else {
    for (var i = 0; i < layouts.length; i++) {
      this.addLayout(layouts[i]);
    }
  }
};
MapModel.prototype.addLayout = function(layout) {
  var layoutData = null;
  if (layout instanceof LayoutData) {
    layoutData = layout;
  } else {
    layoutData = new LayoutData(layout);
  }
  var object = this._layoutsData[layoutData.getId()];
  if (object === undefined) {
    this._layoutsData[layoutData.getId()] = layoutData;
  } else {
    logger.warn("Layout " + layoutData.getId() + " already exists in a model: " + this.getId());
  }
};

MapModel.prototype.addSubmodels = function(submodels) {
  if (submodels !== undefined) {
    for (var i = 0; i < submodels.length; i++) {
      this.addSubmodel(submodels[i]);
    }
  }
};

MapModel.prototype.addSubmodel = function(submodel) {
  if (!(submodel instanceof MapModel)) {
    submodel = new MapModel(submodel);
  }
  this._submodels.push(submodel);
};

MapModel.prototype.getSubmodels = function() {
  return this._submodels;
};

MapModel.prototype.getSubmodelById = function(id) {
  if (this.getId() === id) {
    return this;
  }
  for (var i = 0; i < this._submodels.length; i++) {
    if (this._submodels[i].getId() === id) {
      return this._submodels[i];
    }
  }
  return null;
};

MapModel.prototype._getLayouts = function() {
  var result = [];
  for ( var id in this._layoutsData) {
    if (this._layoutsData.hasOwnProperty(id)) {
      result.push(this._layoutsData[id]);
    }
  }
  return result;
};

MapModel.prototype.getByIdentifiedElement = function(ie, complete) {
  var self = this;
  if (ie.getType() === "ALIAS") {
    return self.getAliasById(ie.getId(), complete);
  } else if (ie.getType() === "REACTION") {
    return self.getReactionById(ie.getId(), complete);
  } else if (ie.getType() === "POINT") {
    var id = self._pointToId(ie.getId());
    var result = this._pointsData[id];
    if (result === undefined) {
      result = new PointData(ie);
      this._pointsData[id] = result;
    }
    return Promise.resolve(result);
  } else {
    throw new Error("Unknown type: " + ie.getType());
  }
};

MapModel.prototype.getByIdentifiedElements = function(identifiedElements, complete) {
  var self = this;
  var missingAliases = [];
  var missingReactions = [];

  for (var i = 0; i < identifiedElements.length; i++) {
    var ie = identifiedElements[i];
    if (!this.isAvailable(ie, complete)) {
      if (ie.getType() === "ALIAS") {
        missingAliases.push(ie.getId());
      } else if (ie.getType() === "REACTION") {
        missingReactions.push(ie.getId());
      } else {
        throw new Error("Unknown type " + ie);
      }
    }
  }

  return self.getMissingElements({
    aliasIds : missingAliases,
    reactionIds : missingReactions,
    complete : complete
  }).then(function() {
    var promises = [];
    for (var i = 0; i < identifiedElements.length; i++) {
      promises.push(self.getByIdentifiedElement(identifiedElements[i], complete));
    }
    return Promise.all(promises);
  });

};

MapModel.prototype.isAvailable = function(ie, complete) {
  var element;
  if (ie.getType() === "ALIAS") {
    element = this._aliases[ie.getId()];
  } else if (ie.getType() === "REACTION") {
    element = this._reactions[ie.getId()];
  } else if (ie.getType() === "POINT") {
    var id = this._pointToId(ie.getId());
    var result = this._pointsData[id];
    if (result === undefined) {
      result = new PointData(ie);
      this._pointsData[id] = result;
    }
    element = this._pointsData[id];
  } else {
    throw new Error("Unknown type: " + ie.getType(), complete);
  }
  if (element === undefined) {
    return false;
  } else if (complete) {
    return element.isComplete();
  } else {
    return true;
  }
};

MapModel.prototype.getReactionsForElement = function(element, complete) {
  return this.getReactionsForElements([ element ], complete);
};

MapModel.prototype.getReactionsForElements = function(elements, complete) {
  var self = this;
  var ids = [];
  var i;
  for (i = 0; i < elements.length; i++) {
    ids.push(elements[i].getId());
  }
  var idString = ids.join();
  if (this._reactionsByParticipantElementId[idString]) {
    var reactions = self._reactionsByParticipantElementId[idString];
    if (!complete) {
      return Promise.resolve(reactions);
    } else {
      var promises = [];
      for (i = 0; i < reactions.length; i++) {
        promises.push(self.getCompleteReactionById(reactions[i].getId()));
      }
      return Promise.all(promises);
    }
  }

  var result = [];
  return ServerConnector.getReactions({
    participantId : ids,
  }).then(function(reactions) {
    result = reactions;

    for (var i = 0; i < reactions.length; i++) {
      var reaction = reactions[i];
      var id = reaction.getId();
      if (self._reactions[id] === undefined) {
        self._reactions[id] = reaction;
      } else {
        self._reactions[id].update(reaction);
      }
    }
    var ids = self._getMissingReactionsElementIds(reactions);
    return self.getMissingElements({
      aliasIds : ids,
      complete : true
    });
  }).then(function() {
    var promises = [];
    for (var i = 0; i < result.length; i++) {
      promises.push(self.getCompleteReactionById(result[i].getId()));
    }
    return Promise.all(promises);
  });
};

MapModel.prototype.getCompartments = function() {
  var self = this;

  var promise = Promise.resolve();
  if (self._compartments === undefined) {
    promise = ServerConnector.getAliases({
      columns : "id,bounds,modelId",
      type : "Compartment",
      modelId : self.getId(),
    }).then(function(compartments) {
      self._compartments = [];
      for (var i = 0; i < compartments.length; i++) {
        self._compartments.push(new IdentifiedElement(compartments[i]));
      }
    });
  }
  return promise.then(function() {
    return self.getByIdentifiedElements(self._compartments, true);
  });
};

module.exports = MapModel;
